﻿local Rule = {
	Item = false,
	Execute = { }, -- rule functions
	Entry = { }, -- database functions
}

function ArkInventory.RuleAppliesToItem( rulenum, i )

	local ra = ArkInventory.db.account.option.rule[rulenum]
	local rp = ArkInventory.db.profile.option.rule[rulenum]
	
	if not i or not ra or ra.damaged or not rp then
		return false, nil
	end
	
	if ra.compiled == ""  then
		return false, "not compiled"
	end
	
	local p, eor = loadstring( ra.compiled, "Error in rule " .. rulenum .. " [" .. ra.name .. "]" )

	if not p then
		return nil, eor
	end
	
	Rule.Item = i
	
	setfenv( p, Rule.Environment )
	ok, eor = pcall( p )
	
	if not ok then
		return nil, eor
	else
		return eor, nil
	end
	
end

function Rule.Execute.Soulbound()

	if Rule.Item.sb then
		return true
	end

	return false
	
end

function Rule.Execute.Empty()

	if not Rule.Item.h or strtrim( Rule.Item.h ) == "" then
		return true
	end

	return false
	
end

function Rule.Execute.ItemId( t )

	if not Rule.Item.h or not t or strtrim( t ) == "" then return false end
	
	for w in t:gmatch( ArkInventory.Localise["RULE_PATTERN_SEP"] ) do
		
		w = strtrim( w ) .. ":"
		e = string.sub( ArkInventory.ItemGetID( Rule.Item.h ) .. ":", 1, string.len( w ) )
	
		if w == e then
			return true
		end
		
	end
	
	return false
	
end

function Rule.Execute.ItemType( t )

	if not Rule.Item.h or not t or strtrim( t ) == "" then return false end
	
	e = string.lower( select( 6, GetItemInfo( Rule.Item.h ) ) or "" )
	
	if e ~= "" then
	
		for w in t:gmatch( ArkInventory.Localise["RULE_PATTERN_SEP"] ) do
		
			if e == string.lower( strtrim( w ) ) then
				return true
			end
			
		end
		
	end
	
	return false
	
end

function Rule.Execute.ItemSubType( t )

	if not Rule.Item.h or not t or strtrim( t ) == "" then return false end
	
	e = string.lower( select( 7, GetItemInfo( Rule.Item.h ) ) or "" )
	
	if e ~= "" then
	
		for w in t:gmatch( ArkInventory.Localise["RULE_PATTERN_SEP"] ) do
		
			if e == string.lower( strtrim( w ) ) then
				return true
			end
			
		end
		
	end
	
	return false
	
end

function Rule.Execute.ItemEquipLocation( t )

	if not Rule.Item.h then return false end
	
	local e = strtrim( select( 9, GetItemInfo( Rule.Item.h ) ) or "" )
	if string.len( e ) > 1 then
		e = getglobal( e )
	end
	e = string.lower( strtrim( e ) )
	
	t = string.lower( strtrim( t ) )
	
	if e ~= "" then
	
		if t == "" then
		
			-- no tokens so just check if item can be equipped
			if e == "" then
				return false
			else
				return true
			end
		
		else
		
			for w in t:gmatch( ArkInventory.Localise["RULE_PATTERN_SEP"] ) do
			
				if e == strtrim( w ) then
					return true
				end
				
			end
			
		end
		
	end
	
	return false
	
end

function Rule.Execute.ItemName( t )

	if not Rule.Item.h or not t or strtrim( t ) == "" then return false end
	
	e = string.lower( ArkInventory.ItemGetName( Rule.Item.h ) or "" )
	t = string.lower( strtrim( t ) )
	
	for w in string.gmatch( t, ArkInventory.Localise["RULE_PATTERN_SEP"] ) do

		if string.find( e, strtrim( w ) ) then
			return true
		end
			
	end
	
	return false
	
end

function Rule.Execute.ItemQuality( t )

	if not Rule.Item.h or not t or strtrim( t ) == "" then return false end
	
	t = string.lower( strtrim( t ) )
	for z = 6, 0, -1 do
		q = string.lower( getglobal( "ITEM_QUALITY" .. z .. "_DESC" ) or tostring( z ) )
		t = t:gsub( q, tostring( z ) )
	end

	for w in t:gmatch( ArkInventory.Localise["RULE_PATTERN_SEP"] ) do
	
		w = strtrim( w )
		
		if w == tostring( Rule.Item.q or 0 ) then
			return true
		end

	end
	
	return false
	
end

function Rule.Execute.ItemLevelStat( t )

	if not Rule.Item.h or not t or strtrim( t ) == "" then return false end
	
	local level = select( 4, GetItemInfo( Rule.Item.h ) )
	
	local t = strtrim( t )

	local imin = 0
	local imax = 0
	
	imin = strmatch( t, "^(%d+)" )
	if not imin then
		error( "invalid argument " .. t )
	end
	
	imax = strmatch( t, imin .. "[^%d]-(%d+)" )

	imin = tonumber( imin ) or 0
	imax = tonumber( imax ) or imin
		
	if level >= imin and level <= imax then
		return true
	end

	return false
	
end

function Rule.Execute.ItemLevelUse( t )

	if not Rule.Item.h or not t or strtrim( t ) == "" then return false end
	
	local level = select( 5, GetItemInfo( Rule.Item.h ) )
	
	local t = strtrim( t )

	local imin = 0
	local imax = 0
	
	imin = strmatch( t, "^(%d+)" )
	if not imin then
		error( "invalid argument " .. t )
	end
	
	imax = strmatch( t, imin .. "[^%d]-(%d+)" )

	imin = tonumber( imin ) or 0
	imax = tonumber( imax ) or imin
		
	if level >= imin and level <= imax then
		return true
	end
	
	return false

end

function Rule.Execute.PeriodicTable( t )

	if not Rule.Item.h or not t or strtrim( t ) == "" then return false end
	
	return ArkInventory.PT_ItemInSets( Rule.Item.h, t )
	
end

function Rule.Execute.Tooltip( t )

	local i = Rule.Item
	
	if not i or not i.h or i.bag_id == nil or i.slot_id == nil or not t or strtrim( t ) == "" then return false end
	
	if i.test_rule then
	
		ArkInventory.Global.ScanTooltip:SetHyperlink( i.h )
	
	else

		-- tooltip is also used offline mode
		loc_id = ArkInventory.BagGetLocation( i.bag_id )
		if ArkInventory.Global.Location[loc_id].offline then
			ArkInventory.Global.ScanTooltip:SetHyperlink( i.h )
		elseif i.bag_id == BANK_CONTAINER then
			ArkInventory.Global.ScanTooltip:SetInventoryItem( BankButtonIDToInvSlotID( i.slot_id ) )
		elseif i.bag_id == KEYRING_CONTAINER then
			ArkInventory.Global.ScanTooltip:SetInventoryItem( KeyRingButtonIDToInvSlotID( i.slot_id ) )
		else
			ArkInventory.Global.ScanTooltip:SetBagItem( i.bag_id, i.slot_id )
		end
		
	end

	t = string.lower( strtrim( t ) )

	for w in t:gmatch( ArkInventory.Localise["RULE_PATTERN_SEP"] ) do

		if ArkInventory.Global.ScanTooltip:Contains( strtrim( w ) ) then
			return true
		end
	
	end
	
	return false

end

function Rule.Execute.Outfit( t )

	if not Rule.Item.h then return false end
	
	if IsAddOnLoaded( "Outfitter" ) and Outfitter.Initialized then
		return Rule.Execute.OutfitOutfitter( t, Rule.Item.h )
	elseif IsAddOnLoaded( "ItemRack" ) then
		return Rule.Execute.OutfitItemRack( t, Rule.Item.h )
	elseif IsAddOnLoaded( "ClosetGnome" ) then
		return Rule.Execute.OutfitClosetGnome( t, Rule.Item.h )
	end
	
	return false

end

function Rule.Execute.OutfitOutfitter( t, h )

	if not h or not IsAddOnLoaded( "Outfitter" ) or not Outfitter.Initialized then
		return false
	end
	
	local ItemInfo = Outfitter_GetItemInfoFromLink( h )
	
	if not ItemInfo then
		return false
	end
	
	--if 1 == 1 then return false end
	
	local Outfits = Outfitter_GetOutfitsUsingItem( ItemInfo )
	
	if not Outfits or next( Outfits ) == nil then return false end
	
	if not t or strtrim( t ) == "" then
		-- nil t means is it in any outfit
		return true
	end
	
	for w in t:gmatch( ArkInventory.Localise["RULE_PATTERN_SEP"] ) do

		for _, o in pairs( Outfits ) do
			if string.lower( strtrim( o.Name ) ) == string.lower( strtrim( w ) ) then
				return true
			end
		end
		
	end
	
	return false

end

function Rule.Execute.OutfitClosetGnome( t, h )

	if not h or not IsAddOnLoaded( "ClosetGnome" ) then
		return false
	end

	local Outfits = { }	
	
	if ClosetGnome.GetOutfitsUsingItem then
	
		Outfits = { ClosetGnome:GetOutfitsUsingItem( h ) }

	else

		for n in pairs( ClosetGnome.db.char.set ) do
			for s, v in pairs( ClosetGnome.db.char.set[n] ) do
				if v == h then
					tinsert( Outfits, strtrim( n ) )
					break
				end
			end
		end
		
	end

	
	if not Outfits or next( Outfits ) == nil then return false end
	
	if not t or strtrim( t ) == "" then
		-- nil t means is it in any outfit
		return true
	end

	for w in t:gmatch( ArkInventory.Localise["RULE_PATTERN_SEP"] ) do

		for _, o in pairs( Outfits ) do
			if string.lower( o ) == string.lower( strtrim( w ) ) then
				return true
			end
		end
	
	end
	
	return false
	
end

function Rule.Execute.OutfitItemRack( t, h )

	if not h or not IsAddOnLoaded( "ItemRack" ) then return false end
	
	id = strmatch( h, "item:(.+):")
	
	local Outfits = { }
	
	for n in pairs( ItemRackUser.Sets ) do
		for _, v in pairs( ItemRackUser.Sets[n].equip ) do
			if id == v and string.sub( n, 1, 1 ) ~= "~" then
				tinsert( Outfits, strtrim( n ) )
				break
			end
		end
	end
	
	if not Outfits or next( Outfits ) == nil then return false end
	
	if not t or strtrim( t ) == "" then
		-- nil t means is it in any outfit
		return true
	end

	for w in t:gmatch( ArkInventory.Localise["RULE_PATTERN_SEP"] ) do

		for _, o in pairs( Outfits ) do
			if string.lower( o ) == string.lower( strtrim( w ) ) then
				return true
			end
		end
	
	end
	
	return false

end

function Rule.Execute.Wearble()

	-- can be equipped
	if not Rule.Execute.ItemEquipLocation() then
		return false
	end

	-- player level high enough to use
	if not Rule.Execute.ItemLevelUse( "0-" .. UnitLevel( "player" ) ) then
		return false
	end

	-- class restrictions
	--ArkInventory.Global.Me.class
	
	-- armor type restriction
	
end

function Rule.Execute.CharacterLevelRange( t )

	-- ( levels below, levels above )

	if not Rule.Item.h or not t or strtrim( t ) == "" then return false end

	local clevel = UnitLevel( "player" )
	local ilevel = select( 5, GetItemInfo( Rule.Item.h ) ) or clevel
	
	local t = strtrim( t )

	local v_below = strmatch( t, "^(%d+)" )
	if not v_below then
		error( "invalid argument " .. t )
	end
	
	local v_above = strmatch( t, v_below .. "[^%d]-(%d+)" )
	if not v_above then
		-- single value
		v_above = v_below
	end

	v_below = clevel - tonumber( v_below )
	v_above = clevel + tonumber( v_above )
	
	if ilevel >= v_below and ilevel <= v_above then
		return true
	end
	
	return false
	
end

Rule.Environment = {
	soulbound=Rule.Execute.Soulbound,
	sb=Rule.Execute.Soulbound,
	empty=Rule.Execute.Empty,
	id=Rule.Execute.ItemId,
	type=Rule.Execute.ItemType,
	itemtype=Rule.Execute.ItemType,
	subtype=Rule.Execute.ItemSubType,
	equip=Rule.Execute.ItemEquipLocation,
	name=Rule.Execute.ItemName,
	quality=Rule.Execute.ItemQuality,
	periodictable=Rule.Execute.PeriodicTable,
	pt=Rule.Execute.PeriodicTable,
	tooltip=Rule.Execute.Tooltip,
	tt=Rule.Execute.Tooltip,
	outfit=Rule.Execute.Outfit,
	itemlevel=Rule.Execute.ItemLevelStat,
	ilvl=Rule.Execute.ItemLevelStat,
	statlevel=Rule.Execute.ItemLevelStat,
	ireq=Rule.Execute.ItemLevelUse,
	uselevel=Rule.Execute.ItemLevelUse,
	clr=Rule.Execute.CharacterLevelRange,
}



function ArkInventory.Frame_Rules_Hide()
	ARKINV_Rules:Hide()
end
	
function ArkInventory.Frame_Rules_Show()
	ARKINV_Rules:Show()
end


function ArkInventory.Frame_Rules_Paint( frame )

	-- scale
	
	-- background
	
	-- border
	
	-- title
	obj = getglobal( frame:GetName() .. "TitleWho" )

	t = string.format( "%s: %s v%s", ArkInventory.Localise["MOD_MENU_MAIN_RULES"], ArkInventory.Const.Program.Name, ArkInventory.Const.Program.UIVersion )

	if ArkInventory.Const.Program.Beta then
		t = t .. " " .. RED_FONT_COLOR_CODE .. "(beta " .. ArkInventory.Const.Program.Beta .. ")"
	end
	
	obj:SetText( t )
	
end

function ArkInventory.UiSetEditBoxLabel( arg1 )
	getglobal( this:GetName() .. "Label" ):SetText( arg1 )
end

function ArkInventory.Frame_Rules_TabToNext( w_curr, w_prev, w_next )

	if not w_curr or not w_prev or not w_next then
		return
	end
	
	local f = this:GetName()
	f = string.sub( f, 1, string.len( f ) - string.len( w_curr ) )
	
	if IsShiftKeyDown() then
		f = f .. w_prev
	else
		f = f .. w_next
	end

	local w = getglobal( f )
	if w then
		w:SetFocus()
	end

end


function ArkInventory.Frame_Rules_Build_Sort()

	local f = this:GetName()
	
	local x
	
	--enabled
	x = getglobal( f .. "_T1" )
	x:ClearAllPoints()
	x:SetWidth( 32 )
	x:SetPoint( "TOP", 0, 0 )
	x:SetPoint( "BOTTOM", 0, 0 )
	x:SetPoint( "LEFT", 15, 0 )
	x:SetText( ArkInventory.Localise["MOD_FRAME_RULE_LIST_ENABLED"] )
	x:Show()

	--damaged
	x = getglobal( f .. "_T2" )
	x:ClearAllPoints()
	x:SetWidth( 32 )
	x:SetPoint( "TOP", 0, 0 )
	x:SetPoint( "BOTTOM", 0, 0 )
	x:SetPoint( "LEFT", f .. "_T1", "RIGHT", 5, 0 )
	x:SetText( ArkInventory.Localise["MOD_FRAME_RULE_LIST_DAMAGED"] )
	x:Show()
	
	-- id
	x = getglobal( f .. "_C1" )
	x:ClearAllPoints()
	x:SetWidth( 50 )
	x:SetPoint( "LEFT", f .. "_T2", "RIGHT", 5, 0 )
	x:SetPoint( "TOP", 0, 0 )
	x:SetPoint( "BOTTOM", 0, 0 )
	x:SetText( ArkInventory.Localise["MOD_FRAME_RULE_LIST_ID"] )
	x:Show()

	-- order
	x = getglobal( f .. "_C2" )
	x:ClearAllPoints()
	x:SetWidth( 50 )
	x:SetPoint( "LEFT", f .. "_C1", "RIGHT", 5, 0 )
	x:SetPoint( "TOP", 0, 0 )
	x:SetPoint( "BOTTOM", 0, 0 )
	x:SetText( ArkInventory.Localise["MOD_FRAME_RULE_LIST_ORDER"] )
	x:Show()

	-- description
	x = getglobal( f .. "_C3" )
	x:ClearAllPoints()
	x:SetPoint( "LEFT", f .. "_C2", "RIGHT", 5, 0 )
	x:SetPoint( "TOP", 0, 0 )
	x:SetPoint( "BOTTOM", 0, 0 )
	x:SetPoint( "RIGHT", -35, 0 )
	x:SetText( ArkInventory.Localise["MOD_FRAME_RULE_LIST_DESCRIPTION"] )
	x:Show()
	
end

function ArkInventory.Frame_Rules_Build_Row()

	local f = this:GetName()
	
	local x
	local sz = 18
	
	--enabled
	x = getglobal( f .. "T1" )
	x:ClearAllPoints()
	x:SetWidth( sz )
	x:SetHeight( sz )
	x:SetPoint( "LEFT", 17, 0 )
	x:Show()

	--damaged
	x = getglobal( f .. "T2" )
	x:ClearAllPoints()
	x:SetWidth( sz )
	x:SetHeight( sz )
	x:SetPoint( "LEFT", f .. "T1", "RIGHT", 19, 0 )
	x:Show()
	
	-- id
	x = getglobal( f .. "C1" )
	x:ClearAllPoints()
	x:SetWidth( 50 )
	x:SetPoint( "LEFT", f .. "T2", "RIGHT", 12, 0 )
	x:SetPoint( "TOP", 0, 0 )
	x:SetPoint( "BOTTOM", 0, 0 )
	x:SetTextColor( 1, 1, 1, 1 )
	x:SetJustifyH( "CENTER", 0, 0 )
	x:Show()

	-- order
	x = getglobal( f .. "C2" )
	x:ClearAllPoints()
	x:SetWidth( 50 )
	x:SetPoint( "LEFT", f .. "C1", "RIGHT", 5, 0 )
	x:SetPoint( "TOP", 0, 0 )
	x:SetPoint( "BOTTOM", 0, 0 )
	x:SetTextColor( 1, 1, 1, 1 )
	x:SetJustifyH( "CENTER", 0, 0 )
	x:Show()

	-- description
	x = getglobal( f .. "C3" )
	x:ClearAllPoints()
	x:SetPoint( "LEFT", f .. "C2", "RIGHT", 5, 0 )
	x:SetPoint( "TOP", 0, 0 )
	x:SetPoint( "BOTTOM", 0, 0 )
	x:SetPoint( "RIGHT", -5, 0 )
	x:SetJustifyH( "LEFT", 0, 0 )
	x:Show()

end

function ArkInventory.Frame_Rules_Build_Table( f )

	if not f or type( f ) ~= "string" then f = this:GetName() end

	local p = getglobal( f )
	if not p then
		ArkInventory.PrintPlus( { RED_FONT_COLOR_CODE, "OOPS: Invalid value at ArkInventory.Frame_Rules_Build_Table( [", f, "] )" } )
		return
	end
	
	local maxrows = tonumber( getglobal( f .. "MaxRows" ):GetText() )
	local rows = maxrows
	local height = 24
	
	if rows > maxrows then rows = maxrows end
	getglobal( f .. "NumRows" ):SetText( rows )
	
	if height == 0 then
		height = tonumber( getglobal( f .. "RowHeight" ):GetText() )
	end
	getglobal( f .. "RowHeight" ):SetText( height )

	-- stretch scrollbar to bottom row
	getglobal( f .. "Scroll" ):SetPoint( "BOTTOM", f .. "Row" .. rows, "BOTTOM", 0, 0 )

	-- set frame height to correct size
	getglobal( f ):SetHeight( height * rows + 20 )

end

function ArkInventory.Frame_Rules_Table_Row_OnClick()

	local f = this:GetName()
	
	-- ArkInventory.Print( "RuleTableClick( " .. f .. " )" )
	local parent = getglobal( f ):GetParent():GetName()
	
	local cs = getglobal( parent .. "SelectedRow" ):GetText()
	local ns = tostring( getglobal( f ):GetID() )

	if ns == "0" then
		ArkInventory.PrintPlus( { "OOPS: widget [", f, "] has no ID allocated" } )
		return false
	end
	
	
	if IsShiftKeyDown() then
	
		-- shift click - enable/disable the rule
		id = tonumber( getglobal( f .. "Id" ):GetText() )
		if id > 0 then
		
			if ArkInventory.db.profile.option.rule[id] then
			
				ArkInventory.db.profile.option.rule[id] = false
				ArkInventory.ItemCacheClear()
				ArkInventory.Frame_Main_Generate_RecalcAll()
			
			else
		
				d = ArkInventory.db.account.option.rule[id]
				d["enabled"] = true
				ArkInventory.RuleEntryEdit( id, d )
				
			end
			
		end
	
		ArkInventory.Frame_Rules_Table_Refresh()
		
	else
	
		-- normal click - show/hide selected background
	
		if cs ~= "-1" then
			getglobal( parent .. "Row" .. cs .. "Selected" ):Hide()
		end

		-- second click removes selection		
		if cs == ns then
			getglobal( parent .. "SelectedRow" ):SetText( "-1" )
			getglobal( parent .. "SelectedId" ):SetText( "-1" )
			return
		end
	
		getglobal( parent .. "SelectedRow" ):SetText( ns )
		getglobal( parent .. "SelectedId" ):SetText( getglobal( f .. "Id" ):GetText() )
	
		getglobal( f .. "Selected" ):Show()
		
	end

end

function ArkInventory.Frame_Rules_Table_Reset( f )

	if not f or type( f ) ~= "string" then f = this:GetName() end
	-- hide and reset all rows
	
	local t = f .. "Table"
	
	local h = tonumber( getglobal( t .. "RowHeight" ):GetText() )
	local r = tonumber( getglobal( t .. "NumRows" ):GetText() )

	getglobal( t .. "SelectedRow" ):SetText( "-1" )
	for x = 1, r do
		getglobal( t .. "Row" .. x .. "Selected" ):Hide()
		getglobal( t .. "Row" .. x .. "Id" ):SetText( "-1" )
		getglobal( t .. "Row" .. x ):Hide()
		getglobal( t .. "Row" .. x ):SetHeight( h )
	end

end

function ArkInventory.Frame_Rules_Table_Refresh( f )

	if not f then
		f = this:GetParent():GetParent():GetParent():GetName()
	end

	-- ArkInventory.Print( "Frame_Rules_Table_Refresh( " .. ArkInventory.nilStringText( f ) .. " ), " .. this:GetName() )

	local p = getglobal( f )
	if not p then
		ArkInventory.PrintPlus( { RED_FONT_COLOR_CODE, "OOPS: Invalid widget name at Frame_Rules_Table_Refresh( ", f, " )" } )
		return
	end
	
	f = f .. "View"
	
	local ft = f .. "Table"
	local fs = f .. "Search"

	local height = tonumber( getglobal( ft .. "RowHeight" ):GetText() )
	local rows = tonumber( getglobal( ft .. "NumRows" ):GetText() )

	local line
	local lineplusoffset
	
	ArkInventory.Frame_Rules_Table_Reset( f )

	-- collect all valid items
	local filter = getglobal( fs .. "Filter"):GetText()
	--ArkInventory.Print( "filter = [" .. filter .. "]" )

	local t = { }
	local tc = 0
	
	local ignore
	
	for k, d in pairs( ArkInventory.db.account.option.rule ) do

		-- ArkInventory.Print( "k = [" .. k .. "], order = [" .. d.order .. "], name = [" .. ArkInventory.nilStringEmpty( d.name ) .. "], formula = [" .. ArkInventory.nilStringEmpty( d.formula ) .. "]" )
	
		ignore = false
		
		if filter ~= "" then
			if not string.find( string.lower( d.name or "" ), string.lower( filter ) ) then
				ignore = true
			end
		end
		
--		if ArkInventory.db.account.option.rule[k].order < 0 then
--			ignore = true
--		end
		
		if not ArkInventory.db.account.option.showdisabled and not ArkInventory.db.profile.option.rule[k] then
			ignore = true
		end
		
		if not ignore then
			table.insert( t, { ["sorted"]=format( "%04i %04i", d.order or 0, k ), ["id"]=k, ["enabled"]=ArkInventory.db.profile.option.rule[k] or false, ["order"]=d.order or 0, ["name"]=d.name or "", ["formula"]=d.formula or "", ["damaged"]=d.damaged or false } )
			tc = tc + 1
		end

	end

	if tc == 0 then
		return
	end
	
	-- sort them by name
	table.sort( t, function( a, b ) return a.sorted < b.sorted end )

	FauxScrollFrame_Update( getglobal( ft .. "Scroll" ), tc, rows, height )
	
	local linename, c, r
	
	for line = 1, rows do

		linename = ft .. "Row" .. line
		
		lineplusoffset = line + FauxScrollFrame_GetOffset( getglobal( ft .. "Scroll" ) )

		if lineplusoffset <= tc then

			c = ""
			r = t[lineplusoffset]
			
			getglobal( linename .. "Id" ):SetText( r.id )

			if r.enabled then
				getglobal( linename .. "T1" ):SetTexture( "Interface\\Icons\\Spell_ChargePositive" )
			else
				getglobal( linename .. "T1" ):SetTexture( "Interface\\Icons\\Spell_ChargeNegative" )
			end
			
			if r.damaged then
				getglobal( linename .. "T2" ):SetTexture( "Interface\\Icons\\Spell_Shadow_DeathCoil" )
			else
				getglobal( linename .. "T2" ):SetTexture( 0, 0, 0, 0 )
			end

			getglobal( linename .. "C1" ):SetText( r.id )

			c = string.format( r.order )
			getglobal( linename .. "C2" ):SetText( c )

			c = r.name
			if not c then c = "<not set>" end
			getglobal( linename .. "C3" ):SetText( c )

			getglobal( linename ):Show()
			
			-- show selected if id is scrolled into view
			if getglobal( ft .. "SelectedId" ):GetText() == r.order then
				getglobal( ft .. "SelectedRow" ):SetText( line )
				getglobal( ft .. "Row" .. line .. "Selected" ):Show()
			end
			
		else
		
			getglobal( linename .. "Id"):SetText( "-1" )
			getglobal( linename ):Hide()
			
		end
	end

	-- ~~~~ clean table t
	
end

function ArkInventory_Frame_Rules_Table_Refresh( f )
	ArkInventory.Frame_Rules_Table_Refresh( f )
end



function ArkInventory.RuleEntryCompile( d )

	d["compiled"] = ""
	
	if not d["formula"] then
		return
	end
	
	d.compiled = "return ( " .. d.formula .. " );"
	d.compiled = string.gsub( d.compiled, "[\r]", " " )
	d.compiled = string.gsub( d.compiled, "[\n]", " " )
	d.compiled = string.gsub( d.compiled, "%s+", " " )
	d.compiled = string.gsub( d.compiled, "(%a+)%((.-)%)", '%1("%2")' )
	
end

function ArkInventory.RuleEntryValid( d )

	if not d.formula or strtrim( d.formula ) == "" then
		return false, "Formula is empty"
	end
	
	ArkInventory.RuleEntryCompile( d )
	
	Rule.Item = { ["test_rule"]=true, ["bag_id"]=0, ["slot_id"]=1, ["count"]=1, ["q"]=1, ["sb"]=true, ["h"]="item:6948:0:0:0:0:0:0:0" } -- hearthstone
	
	local p, eor = loadstring( d.compiled, "Error in formula" )
	
	if not p then
		return false, eor
	end
	
	setfenv( p, Rule.Environment )
	ok, eor = pcall( p )
	
	if not ok then
		return false, eor
	else
		return true
	end
	
end

function ArkInventory.RuleEntryFormat( d )

	if not d then
		return
	end
	
	zOrder = 9999
	zName = ""
	zFormula = ""

	local x1 = tonumber( d["order"] )

	if x1 then
		zOrder = x1
	end

	if zOrder < 0 then
		zOrder = 9999
	end
	
	if zOrder > 9999 then
		zOrder = 9999
	end
	
	local x1 = d["name"]
	if x1 then
		zName = x1
	end

	local x1 = d["formula"]
	if x1 then
		zFormula = x1
	end

	d["damaged"] = false
	d["order"] = tonumber( zOrder )
	d["name"] = strtrim( tostring( zName ) )
	d["formula"] = string.gsub( strtrim( tostring( zFormula ) ), "[%a]-%(", string.lower )
	d.formula = string.gsub( d.formula, "%s+", " " )

	if d["enabled"] and ( d.enabled == 1 or d.enabled == true ) then
		d["enabled"] = true
	else
		d["enabled"] = false
	end

	return d
	
end

function ArkInventory.RuleEntryUpdate( k, d )

	k = tonumber( k )
	ArkInventory.RuleEntryFormat( d )

	-- increment the rule counter if this is a new rule
	if not ArkInventory.RuleEntryExists( k ) then
		ArkInventory.db.account.option.nextrule = ArkInventory.db.account.option.nextrule + 1
	end

	-- enable/disable the rule at the profile level
	if d.enabled then
		ArkInventory.db.profile.option.rule[k] = true
	else
		ArkInventory.db.profile.option.rule[k] = nil
	end
	d.enabled = nil
	
	-- save the rule data at the account level
	ArkInventory.db.account.option.rule[k] = d
	
	ArkInventory.ItemCacheClear()
	ArkInventory.Frame_Main_Generate_RecalcAll()
	
end

function ArkInventory.RuleEntryValidate( k, d )

	if not k then
		return false, ArkInventory.Localise["RULE_FAILED_NIL_KEY"]
	end
	
	if not d then
		return false, ArkInventory.Localise["RULE_FAILED_NIL_DATA"]
	end

	ArkInventory.RuleEntryFormat( d )
	
	if d.name == "" then
		return false, ArkInventory.Localise["RULE_FAILED_NIL_DESCRIPTION"]
	end
	
	if d.formula == "" then
		return false, ArkInventory.Localise["RULE_FAILED_NIL_FORMULA"]
	end
	
	ok, em = ArkInventory.RuleEntryValid( d )
	if not ok then
		d["damaged"] = true
		return false, em
	end

	return true

end

function ArkInventory.RuleEntryExists( k )

	if not k then
		error( "passed argument is nil" )
	end
	
	if ArkInventory.db.account.option.rule[tonumber( k )] then
		return true
	else
		return false
	end

end

function ArkInventory.RuleEntryAdd( k, d )
	
	local ok, msg = ArkInventory.RuleEntryValidate( k, d )
	if not ok then
		message( msg )
		return false
	end
	
	ArkInventory.RuleEntryUpdate( k, d )

	return true

end

function ArkInventory.RuleEntryEdit( k, d )

	local ok, msg = ArkInventory.RuleEntryValidate( k, d )
	if not ok then
		message( msg )
		return false
	end
	
	ArkInventory.RuleEntryUpdate( k, d )
	
	return true
	
end

function ArkInventory.RuleEntryRemove( k )

	if not k then
		error( "FAILED: key is nil" )
	end
	
	for k, v in pairs( ArkInventory.db.account.option.rule[tonumber( k )] ) do
		v = nil
	end
	ArkInventory.db.account.option.rule[tonumber( k )] = nil

	ArkInventory.ItemCacheClear()
	ArkInventory.Frame_Main_Generate_RecalcAll()
	
	return true
	
end


function ArkInventory.Frame_Rules_Button_Modify( t )

	local f = this:GetParent():GetParent():GetParent():GetName()
	
	local fvt = f .. "ViewTable"

	local fm = f .. "Modify"
	local fmt = fm .. "Title"
	local fmd = fm .. "Data"

	getglobal( fm .. "Type" ):SetText( t )

	local k = getglobal( fvt .. "SelectedId" ):GetText()
	if not k then k = "-1" end
	if t == "A" then k = "-1" end

	local v
	
	if k ~= "-1" then
		local d = ArkInventory.db.account.option.rule[tonumber( k )]
		d["enabled"] = ArkInventory.db.profile.option.rule[tonumber( k )]
		getglobal( fmd .. "Id" ):SetText( k )
		getglobal( fmd .. "Enabled" ):SetChecked( d.enabled )
		getglobal( fmd .. "Order" ):SetText( ArkInventory.nilStringEmpty( d.order ) )
		getglobal( fmd .. "Description" ):SetText( ArkInventory.nilStringEmpty( d.name ) )
		getglobal( fmd .. "ScrollFormula" ):SetText( ArkInventory.nilStringEmpty( d.formula ) )
	else
		getglobal( fmd .. "Id" ):SetText( ArkInventory.db.account.option.nextrule )
		getglobal( fmd .. "Enabled" ):SetChecked( true )
		getglobal( fmd .. "Order" ):SetText( "0" )
		getglobal( fmd .. "Description" ):SetText( "" )
		getglobal( fmd .. "ScrollFormula" ):SetText( "false" )
	end

	getglobal( fmd .. "IdLabel" ):SetText( ArkInventory.Localise["MOD_FRAME_RULE_RULE"] .. ":"  )
	getglobal( fmd .. "EnabledLabel" ):SetText( ArkInventory.Localise["MOD_FRAME_RULE_ENABLED"] .. ":"  )
	getglobal( fmd .. "OrderLabel" ):SetText( ArkInventory.Localise["MOD_FRAME_RULE_ORDER"] .. ":"  )
	getglobal( fmd .. "DescriptionLabel" ):SetText( ArkInventory.Localise["MOD_FRAME_RULE_DESCRIPTION"] .. ":"  )
	getglobal( fmd .. "FormulaLabel" ):SetText( ArkInventory.Localise["MOD_FRAME_RULE_FORMULA"] .. ":" )
	
	getglobal( fmd .. "Enabled" ):Show()
	getglobal( fmd .. "Order" ):Show()
	getglobal( fmd .. "Description" ):Show()
	getglobal( fmd .. "ScrollFormula" ):Show()

	getglobal( fmd .. "EnabledReadOnly" ):SetChecked( getglobal( fmd .. "Enabled" ):GetChecked() )
	getglobal( fmd .. "EnabledReadOnly" ):Hide()
	getglobal( fmd .. "OrderReadOnly" ):SetText( getglobal( fmd .. "Order" ):GetText() )
	getglobal( fmd .. "OrderReadOnly" ):Hide()
	getglobal( fmd .. "DescriptionReadOnly" ):SetText( getglobal( fmd .. "Description" ):GetText() )
	getglobal( fmd .. "DescriptionReadOnly" ):Hide()
	getglobal( fmd .. "FormulaReadOnly" ):SetText( getglobal( fmd .. "ScrollFormula" ):GetText() )
	getglobal( fmd .. "FormulaReadOnly" ):Hide()

	if t == "R" then

		if k == "-1" then return end

		getglobal( fmt .. "Text" ):SetText( string.upper( ArkInventory.Localise["MOD_FRAME_RULE_REMOVE"] ) )

		getglobal( fmd .. "Enabled" ):Hide()
		getglobal( fmd .. "EnabledReadOnly" ):Disable()
		getglobal( fmd .. "EnabledReadOnly" ):Show()

		getglobal( fmd .. "Order" ):Hide()
		getglobal( fmd .. "OrderReadOnly" ):Show()

		getglobal( fmd .. "Description" ):Hide()
		getglobal( fmd .. "DescriptionReadOnly" ):Show()

		getglobal( fmd .. "ScrollFormula" ):Hide()
		getglobal( fmd .. "FormulaReadOnly" ):Show()

	elseif t == "E" then

		if k == "-1" then return end

		getglobal( fmt .. "Text" ):SetText( string.upper( ArkInventory.Localise["MOD_FRAME_RULE_EDIT"] ) )

	elseif t == "A" then

		getglobal( fmt .. "Text" ):SetText( string.upper( ArkInventory.Localise["MOD_FRAME_RULE_ADD"] ) )

	else
		ArkInventory.PrintPlus( { RED_FONT_COLOR_CODE, "OOPS: Uncoded argument ArkInventory.Frame_Rules_Button_Modify( ", t, " )" } )
		return
	end

	getglobal( f .. "View" ):Hide()
	getglobal( fm ):Show()

end

function ArkInventory.Frame_Rules_Button_Modify_Ok()

	local f = this:GetParent():GetParent():GetParent():GetParent():GetName()
	local fm = this:GetParent():GetParent():GetName()
	local fmd = fm .. "Data"
	
	local d = { }
	d["enabled"] = getglobal( fmd .. "Enabled" ):GetChecked()
	d["order"] = getglobal( fmd .. "Order" ):GetText()
	d["name"] = getglobal( fmd .. "Description" ):GetText()
	d["formula"] = getglobal( fmd .. "ScrollFormula" ):GetText()

	local k = getglobal( fmd .. "Id" ):GetText()

	f = this:GetParent():GetParent():GetParent():GetName()
	fm = this:GetParent():GetParent():GetName()
	
	local t = getglobal( fm .. "Type" ):GetText()

	if t =="A" then
		if not ArkInventory.RuleEntryAdd( k, d ) then return end
		getglobal( f .. "ViewTableSelectedId" ):SetText( k )
	elseif t == "E" then
		if not ArkInventory.RuleEntryEdit( k, d, false ) then return end
	elseif t == "R" then
		if not ArkInventory.RuleEntryRemove( k ) then return end
		getglobal( f .. "ViewTableSelectedId" ):SetText( "-1" )
	else
		ArkInventory.PrintPlus( { RED_FONT_COLOR_CODE, "OOPS: Uncoded value [", t, "] at ArkInventory.Frame_Rules_Button_Modify_Ok" } )
		return
	end

	getglobal( fm ):Hide()
	getglobal( f .. "View" ):Show()

end

function ArkInventory.Frame_Rules_Button_Modify_Cancel()

	f = this:GetParent():GetParent():GetParent():GetName()
	
	getglobal( f .. "Modify" ):Hide()
	getglobal( f .. "View" ):Show()

end


function ArkInventory.Frame_Rules_Button_View_Add()
	return ArkInventory.Frame_Rules_Button_Modify( "A" )
end

function ArkInventory.Frame_Rules_Button_View_Edit()
	return ArkInventory.Frame_Rules_Button_Modify( "E" )
end

function ArkInventory.Frame_Rules_Button_View_Remove()
	return ArkInventory.Frame_Rules_Button_Modify( "R" )
end




--[[

function Rule.Execute.BagNumber( t )

	if not t or not n then return "false" end

	t = string.lower( strtrim( t ) )
	
	ArkInventory.Print( "token=[" .. t .. "], bag=[" .. n .. "]" )

	if t == "bag" or t == ArkInventory.Localise["RULE_PATTERN_BAGNUM_BAG"] then
		t = table.concat( ArkInventory.Global.Location[ArkInventory.Const.Location.Bag].BAGS, "," )
	elseif t == "bank" or t == ArkInventory.Localise["RULE_PATTERN_BAGNUM_BANK"] then
		t = table.concat( ArkInventory.Global.Location[ArkInventory.Const.Location.Bank].BAGS, "," )
	elseif t == "key" or t == ArkInventory.Localise["RULE_PATTERN_BAGNUM_KEY"] then
		t = table.concat( ArkInventory.Global.Location[ArkInventory.Const.Location.Key].BAGS, "," )
	end
		
	for w in t:gmatch( ArkInventory.Localise["RULE_PATTERN_SEP"] ) do
		if n == tonumber( w ) then
			return "true"
		end
	end
	
	return "false"
	
end

function Rule.Execute.SlotNumber( t )

	if not t or not n then return "false" end

	for w in t:gmatch( ArkInventory.Localise["RULE_PATTERN_SEP"] ) do
		if n == tonumber( w ) then
			return "true"
		end
	end
	
	return "false"
	
end

function Rule.Execute.BagType( t )

	if not t or not b then return "false" end

	bt = ArkInventory.BagGetType( b )
	
	for w in t:gmatch( ArkInventory.Localise["RULE_PATTERN_SEP"] ) do

		for k, v in pairs( ArkInventory.Const.Slot.Data ) do
			if v["type"] and string.lower( v.type ) == string.lower( strtrim( w ) ) then
				if bt == k then
					return "true"
				end
			end
		end
		
	end
	
	return "false"
	
end

]]--

